/*-
 * dis - a simple disassembler.
 */
#ifdef __USAGE
%C	[-b brtab] [-v] prog...
where
 -b brtab     use branch-target map brtab to aid disassembly
 -v           verbose; output addresses and breakpoint information.
#endif


#include <libc.h>
#include "analyze.h"
#include "dbi.h"

int verbose = 0;


#define br_flag      br_user
#define br_isflagged br_isuser

put_data(uchar_t *d, unsigned n)
{
	unsigned    nprint = 0;
	unsigned    i;
	while (n--) {
		printf("%s 0x%x", nprint?",":".byte", *d++);
		if (++nprint == 8) {
			printf("\n");
			nprint = 0;
		}
	}
	if (nprint)
		printf("\n");
}

dis(char *src, BreakTab *b)
{
	FILE      *fp;
	lmfo      *lo;
	symiter    iter;
	gbl_info  *g, **ltab;
	uchar_t   *text;
	ulong_t    offs, vbase;

	int        state = 0;

	if ((fp=fopen(src,"r")) == 0) {
		return -1;
	}
	if ((lo=lmf_open(fp)) == 0) {
		return -1;
	}
	if (lmf_loadtext(lo, &text) == -1) {
		lmf_close(lo);
		return -1;
	}
	vbase = lo->code_offset;

	if ((ltab = calloc(lo->code_size, sizeof *ltab)) == 0) {
		free(text);
		lmf_close(lo);
		return -1;
	}

	lmf_symiter(lo, &iter);
	while (g=lmf_symnext(&iter)) {
		if (g->kind & LMF_CODE && g->name[strlen(g->name)-1] == '_')
			ltab[g->addr.offset] = g;
	}

	for (offs=0; offs < lo->code_size && ltab[offs] == 0; offs++);
	while (offs < lo->code_size) {
		char       oinst[100];
		int        n;
		int        i;

/*-
 * if we have a branch table, removed un-reached spots.
 */

		if (b && !br_isvisit(b, offs)) {
			for (n=0; n+offs < lo->code_size; n++) {
				if (br_isvisit(b, offs+n)  ||
				    br_ismarked(b, offs+n) ||
				    ltab[offs+n])
					fprintf(stderr, "warning: unvisited mark at 0x%lx\n", offs);
					break;
			}
			if (n)
				put_data(text+offs, n);
			offs += n;
		}

		n = i386dis(vbase+offs, text+offs, oinst, 1);
/*-
 * look for symbols which would appear in the middle of an opcode
 */
		for (i=1; i < n && !ltab[offs+i]; i++);

		if (i==n && b) {
/*-
 * look for intervening branch targets
 */
			for (i=1; i < n && !br_ismarked(b,offs+i); i++);
		}

		if (i != n) {
/*-
 * one of the above failed -- show it as data.
 */
 			put_data(text+offs, i);
 			offs += i;
 			continue;
		}

		if (ltab[offs]) {
			printf("\n%s:\n", ltab[offs]->name);
			ltab[offs] = 0;
		}

		if (verbose) {
			int      i;
			printf("%08.8x ", vbase+offs);
#if 0
			for (i=0; i < n; i++) {
				printf("%02.2x ", text[offs+i]);
			}
			while (i++ < 14) {
				printf("   ");
			}
#endif
		}
		printf(" %s", oinst);
		if (b && br_ismarked(b, offs)) {
			if (verbose)
				printf(" / breakpoint 0x%x", vbase+offs);
			br_flag(b, offs);
		}
		putchar('\n');
		offs += n;
	}
	if (b){
		for (offs=0; offs < b->bh.bsize; offs++) {
			if (br_ismarked(b,offs) && !br_isflagged(b,offs))
				printf("/ breakpoint 0x%x missed\n", vbase+offs);
		}
	}
	for (offs=0; offs < lo->code_size; offs++) {
		if (ltab[offs]) {
			printf("/ symbol '%s' at 0x%x missed\n",
				ltab[offs]->name, ltab[offs]->addr.offset);
		}
	}
	free(text);
	free(ltab);
	lmf_close(lo);
	return 0;
}

main(int argc, char **argv)
{
	int	c, i;
	BreakTab *b=0;
	while ((c=getopt(argc, argv, "b:v")) != -1) {
		switch (c) {
		case 'v': verbose = 1; break;
		case 'b':
			if ((b=get_brtab(optarg)) == 0) {
				fprintf(stderr, "no breaktab '%s'\n", optarg);
			}
			break;
		}
	}
	for (i=optind; i < argc; i++) {
		if (dis(argv[i], b) == -1) {
			fprintf(stderr, "%s: couldn't disassemble %s\n",
				argv[0], argv[i]);
		}
	}
	return 0;
}
