/*-
 * generate a profile
 */

#include <libc.h>
#include "analyze.h"
#include "ldfile.h"

#include <process.h>
#include <sys/sched.h>
#include <sys/proc_msg.h>
#include <sys/debug.h>
#include <sys/kernel.h>



int verbose;

debug_program(char *path, char **argv)
{
	return qnx_spawn(0, 0, 0, -1, -1,
		_SPAWN_HOLD,
		path, argv, environ, 0, 0);
}

setmem(pid_t tar, uchar_t *m, unsigned n, ulong_t offs)
{
	if (verbose)
		printf("setting address 0x%x to 0x%x\n", offs, *m);
	if (__qnx_debug_xfer(PROC_PID, tar, _DEBUG_MEM_WR,
				m, n, offs, 0x4) == -1) {
		fprintf(stderr, "xfer from 0x%x to 0x%x for 0x%x failed (%s)\n",
			m, offs, n, strerror(errno));
		return -1;
	}
	return 0;
}


__qnx_debug_info(pid_t procpid, pid_t target, struct _debug_info *i)
{
	struct {
		struct _debug_hdr h;
		struct _debug_info i;
	} m;
	struct {
		struct _debug_hdr_reply h;
		struct _debug_info i;
	} r;
	m.h.type = _PROC_DEBUG;
	m.h.subtype = _DEBUG_INFO;
	m.h.pid = target;
	m.h.zero1 = 0;
	m.h.segment = 0;
	m.h.size = 0;
	m.h.offset = 0;
	Send(PROC_PID, &m, &r, sizeof m, sizeof r);
	if (r.h.status == EOK) {
		*i = r.i;
		return 0;
	} else {
		errno = r.h.status;
		return -1;
	}
}



pcontinue(pid_t p)
{
	__qnx_debug_cont(PROC_PID, p, 0);
}


pkill(pid_t p)
{
	 kill(p, SIGKILL);
	__qnx_debug_detach(PROC_PID, p);
	 kill(p, SIGCONT);
}

collect_breaks(pid_t target, pid_t dbpr, BreakTab *b, uchar_t *image)
{
	ulong_t    base = b->bh.base;
	pid_t        p;
	struct _debug_info info;
	struct _reg_struct r;
	int	i;

	if (__qnx_debug_attach(PROC_PID, target, dbpr) == -1) {
		pkill(target);
		return -1;
	}

	for (i=0; i < b->bh.bsize; i++) {
		char    bop = 0xcc;
		if (br_ismarked(b,i)) {
			if (setmem(target, &bop, sizeof bop, i+base) == -1) {
				fprintf(stderr, "can't set breakpt\n");
			}
		}
	}
	pcontinue(target);

	while (1) {
		if ((p=Receive(0,0,0)) != dbpr)
			continue;
		if (__qnx_debug_info(PROC_PID, target, &info)) {
			fprintf(stderr, "debug info failed!?!\n");
			continue;
		}
		switch (info.debug_state) {
		case _DEBUG_STATE_CONT:
			fprintf(stderr, "sitting in CONT -- something fishy\n");
			pkill(target);
			break;
		case _DEBUG_STATE_HELD:
			fprintf(stderr, "process is stopped -- something fishy\n");
			pkill(target);
			break;
		case _DEBUG_STATE_TRACE:
			fprintf(stderr, "stopped by trace count -- something fishy\n");
			pkill(target);
			break;
		case _DEBUG_STATE_BRK:
			if (__qnx_debug_xfer(PROC_PID, target, _DEBUG_REG_RD,
		    			&r, sizeof r, 0, 0) != 0) {
				fprintf(stderr, "xfer registers failed!\n");
				break;
			}
			if (!br_ismarked(b, r.ip-base)) {
				fprintf(stderr, "break point not set?!?\n");
				break;
			}
			br_unmark(b, r.ip-base);
			if (setmem(target, image+r.ip-base, 1, r.ip) == -1) {
				fprintf(stderr, "cannot remove break point\n");
			}
			pcontinue(target);
			break;
		case _DEBUG_STATE_WATCH:
			fprintf(stderr, "watch point tripped -- something fishy\n");
			pkill(target);
			break;
		case _DEBUG_STATE_DEAD:
			__qnx_debug_detach(PROC_PID, target);
			return 0;
		case _DEBUG_STATE_SIGNAL:
			fprintf(stderr, "stopped at signal %d\n", info.signo);
			if (info.signo == SIGSEGV) {
				exit(1);
				pkill(target);
			} else {
				exit(1);
				pcontinue(target);
			}
			break;
		case _DEBUG_STATE_FORK:
		case _DEBUG_STATE_THREAD:
			fprintf(stderr, "forked -- not handled\n");
			pkill(target);
			break;
		default:
			fprintf(stderr, "unknown debug state %x\n",
				info.debug_state);
		}
	}
}

main(int argc, char **argv)
{
	pid_t       target;
	pid_t       dbpr;
	BreakTab   *brtab;
	uchar_t    *brcov=0;
	char       *path;
	uchar_t    *image;
	int         c, i;
	int	    fd;

	char	oname[PATH_MAX];

	oname[0] = 0;

	while ((c=getopt(argc, argv, "vb:o:")) != -1) {
		switch (c) {
		case 'b':
			if ((brtab=get_brtab(optarg)) == 0) {
				fprintf(stderr, "%s: can't open %s\n", argv[0], optarg);
				return 1;
			}
			strcpy(oname, optarg);
			path = brtab->bh.image;
			break;
		case 'o':
			strcpy(oname, optarg);
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			break;
		}
	}
	if (!brtab) {
		return 1;
	}
	if (ldf_loadtext(path, &image) == -1) {
		return 1;
	}
	if ((dbpr=qnx_proxy_attach(0, 0, 0, -1)) == -1)
		return 1;

	target = debug_program(path, argv+optind);
	collect_breaks(target, dbpr, brtab, image);
	put_brtab(oname, brtab);
	if (verbose)
		for (i=0; i < brtab->bh.bsize; i++) {
			if (br_ismarked(brtab, i))
				printf("0x%lx\n", brtab->bh.base+i);
		}
	return 0;
}
