/*-
 * braflow:
 *
 *    analyze a the control flow of a program region; marking addresses
 *    which are branch targets.
 *    The algorithm is naturally recursive, however a non recursive
 *    version is implemented as the 'stack trace' may be of use for
 *    other analysis procedures.
 *
 */

#include <libc.h>
#include "analyze.h"
#include "ldfile.h"

extern int verbose;

typedef struct {
	int    nelem;
	int    sp;
	ulong_t   *stk;
} ulstack;

#define UL_PUSH(isp, e) \
	((isp)->sp < (isp)->nelem ? \
	             ((isp)->stk[(isp)->sp++]=(e),0) : ul_push((isp),(e)))
#define UL_POP(isp, ep) \
	((isp)->sp ? (*(ep) = (isp)->stk[--(isp)->sp], 0) : -1)


static int
ul_initstk(ulstack *isp, int n)
{
	if ((isp->stk=calloc(n, sizeof *isp->stk)) == 0) {
		return -1;
	}
	isp->nelem = n;
	isp->sp    = 0;
	return 0;
}

static int
ul_delstk(ulstack *isp)
{
	if (isp->stk)
		free(isp->stk);
	isp->stk = 0;
	return 0;
}

static int
ul_push(ulstack *isp, ulong_t e)
{
	if (isp->sp >= isp->nelem) {
		ulong_t *x;
		if ((x=realloc(isp->stk, (isp->nelem+256)*sizeof *x)) == 0)
			return -1;
		isp->stk = x;
		isp->nelem += 256;
	}
	isp->stk[isp->sp++] = e;
	return 0;
}
static int
ul_pop(ulstack *isp, ulong_t *ep)
{
	return UL_POP(isp, ep);
}

static
bradbg(unsigned levels, char *fmt, ...)
{
	va_list  v;
	while (levels--)
		putc(' ', stdout);
	va_start(v, fmt);
	vfprintf(stdout, fmt, v);
	va_end(v);
}


int
braflow(text, b, start, lower, upper)
	uchar_t  *text;
	BreakTab *b;
	ulong_t   start, lower, upper;
{
#define TADDR(o, f)  ((f)->offset + (((f)->flags & _OPR_ISREL) ? (o) : -b->bh.base))
	OpCode         op;
	ulstack        stk;
	ulong_t        cur, next;
	int            isvalid = 0;

	ul_initstk(&stk, 256);
	UL_PUSH(&stk, start);

	while (1) {
		if (!isvalid) {
			if (UL_POP(&stk, &cur) == -1) {
				break;	/* stack empty, done */
			}
			if (cur < lower || cur >= upper) {
				if (verbose)
					bradbg(stk.sp, "discard 0x%x\n", cur);
				continue;
			}
			if (br_isvisit(b, cur))
				continue;
			if (verbose) {
				bradbg(stk.sp, "mark 0x%x\n", cur);
			}
			br_visit(b, cur);
			isvalid = 1;
		} else if (cur < lower || cur >= upper) {
			if (verbose)
				bradbg(stk.sp, "discard 0x%x\n", cur);
			isvalid = 0;
			continue;
		} else if (br_isvisit(b, cur)) {
			isvalid = 0;
			continue;
		}
/*-
 * although potentially a valid opcode, it is ver unlikely...
 * It essentially means 'mov [eax], al' or some minor variant;
 * typically storing the bottom bits of a pointer at the location
 * the pointer refers to isn't terribly useful...
 *
 */
		if (text[cur] == 0) {
			cur++;
			continue;
		}
		br_visit(b, cur);
		analyze_opcode(text+cur, &op);
		switch (op.nfol) {
		case 0:  /* TERMINAL INSTRUCTION */
			if (verbose) {
				bradbg(stk.sp, "terminal at 0x%x\n", cur);
			}
			br_term(b, cur);
			isvalid = 0;
			break;
		case 1: /* either unconditional branch or no-control-flow */
			if (op.flags & _OP_ISJMP) {
				next = TADDR(cur, op.foll+0);
				if (verbose) {
					bradbg(stk.sp, "jump to 0x%x [0x%x]\n",
						next, cur);
				}
				br_term(b, cur);
				br_mark(b, next);
				UL_PUSH(&stk, next);
				isvalid = 0;
			} else {
				cur = TADDR(cur, op.foll+0);
			}
			break;
		case 2: /* conditional branch or subroutine call */
			if (verbose) {
				bradbg(stk.sp, "adding target at 0x%x [0x%x]\n",
						TADDR(cur,op.foll+1), cur);
			}
			next = TADDR(cur, op.foll+0);
			br_mark(b, next);
			UL_PUSH(&stk, next);

			next = TADDR(cur, op.foll+1);
			br_mark(b, next);
			UL_PUSH(&stk, next);
			isvalid = 0;
			break;
		}
	}
	ul_delstk(&stk);
	return 0;
}
