#include "dbi.h"
#include <errno.h>
#include <string.h>
#include <malloc.h>

int
lmf_fixname(unsigned char *name)
{
	unsigned len = *name;
	unsigned i;
	for (i=0; i < len; i++)
		name[i] = name[i+1];
	name[i] = '\0';
	return i;
}

/*-
 * generate more useful language table
 */

static
langtab(lmfo *lo, unsigned lsiz)
{
	unsigned     n;
	char        *s = lo->dbase, *es = s + lsiz;

	for (n=1; s < es; n++)
		s += strlen(s)+1;
	if ((lo->langtab = calloc(n+1, sizeof *lo->langtab)) == 0)
		return -1;
	lo->nlang = n;
	s = lo->dbase;
	for (n=0; s < es; n++) {
		lo->langtab[n] = s;
		s += strlen(s)+1;
	}
	return 0;
}

/*-
 * generate more useful module table
 */
static
modtab(lmfo *lo)
{
	mod_info          *m0, *me;
	unsigned           n;
	me = (mod_info *)((char *)lo->sect + lo->sect->gbl_offset);
	m0 = (mod_info *)((char *)lo->sect + lo->sect->mod_offset);
	for (n=0; m0 < me; n++) {
#define NEXT_MODULE(mo) (mod_info *) ((mo)->name+*(mo)->name+1);
		m0 = NEXT_MODULE(m0);
	}
	if ((lo->modtab = calloc(n+1, sizeof *lo->modtab)) == 0) {
		return -1;
	}
	lo->nmods = n;
	m0 = (mod_info *)((char *)lo->sect + lo->sect->mod_offset);
	for (n=0; n < lo->nmods; n++) {
		mod_info  *tmp;
		lo->modtab[n] = m0;
		tmp = NEXT_MODULE(m0);
		lmf_fixname(m0->name);
		m0 = tmp;
	}
	return 0;
#undef NEXT_MODULE
}

static
globtab(lmfo *lo)
{
	gbl_info          *g0, *ge;
	unsigned           n;

	ge = (gbl_info *)((char *)lo->sect + lo->sect->addr_offset);
	g0 = (gbl_info *)((char *)lo->sect + lo->sect->gbl_offset);
	for (n=0; g0 < ge; n++) {
#define NEXT_GLOBAL(mo) (gbl_info *) ((g0)->name+*(g0)->name+1);
		g0 = NEXT_GLOBAL(g0);
	}
	if ((lo->globtab = calloc(n+1, sizeof *lo->globtab)) == 0) {
		return -1;
	}
	lo->nglobs = n;
	g0 = (gbl_info *)((char *)lo->sect + lo->sect->gbl_offset);
	for (n=0; n < lo->nglobs; n++) {
		gbl_info  *tmp;
		lo->globtab[n] = g0;
		tmp = NEXT_GLOBAL(g0);
		lmf_fixname(g0->name);
		g0 = tmp;
	}
#undef NEXT_GLOBAL
	return 0;
}

static
segtab(lmfo *lo)
{
	unsigned           n;
	seg_info          *s0, *se;

	s0 = (seg_info *)((char *)lo->sect+lo->sect->addr_offset);
	se = (seg_info *)((char *)lo->sect+lo->sect->section_size);
	for (n=0; s0 < se; n++) {
		s0 = (seg_info *)(s0->sects+s0->num);
	}
	if ((lo->segtab = calloc(n+1, sizeof *lo->segtab)) == 0) {
		return -1;
	}
	lo->nsegs = n;
	s0 = (seg_info *)((char *)lo->sect+lo->sect->addr_offset);
	for (n=0; n < lo->nsegs; n++) {
		lo->segtab[n] = s0;
		s0 = (seg_info *)(s0->sects+s0->num);
	}
	return 0;
}

lmfo *
lmf_open(FILE *fp)
{
	struct {
		struct _lmf_header h;
		struct _lmf_definition d;
		ulong_t   segs[2];
	}                          l;
	m_dbh m;
	lmfo                       *lo;

	fread(&l, sizeof l, 1, fp);
	fseek(fp, -(long)sizeof m, SEEK_END);
	fread(&m, sizeof m, 1, fp);
	if (m.signature != 0x8386 || m.exe_major_ver > 3) {
		errno = EPROGMISMATCH;
		return 0;
	}
	if ((lo=calloc(1,sizeof *lo)) == 0) {
		errno = ENOMEM;
		return 0;
	}
	lo->fp = fp;
	lo->code_offset = l.d.unmapped_size + l.d.flat_offset;
	lo->code_size   = l.segs[0] & 0xfffffff;
	lo->data_offset = lo->code_offset + l.segs[0] + 0xfff & ~0xfff;
	lo->data_size   = l.segs[1];
	lo->bss_size    = l.d.heap_nbytes;
	lo->stack_size  = l.d.stack_nbytes;

	if ((lo->dbase = malloc(m.debug_size)) == 0) {
		free(lo);
		errno = ENOMEM;
		return 0;
	}
	lo->magic = LMFO_MAGIC;
	fseek(lo->fp, -m.debug_size, SEEK_END);
	fread(lo->dbase, m.debug_size, 1, lo->fp);
	lo->sect = (sect_dbh *)(lo->dbase+m.lang_size+m.segment_size);
	if ((char *)lo->sect+lo->sect->section_size != 
		lo->dbase+m.debug_size-sizeof m) {
		lmf_close(lo);
		errno = EPROGMISMATCH;
		return 0;
	}

	if (langtab(lo, m.lang_size) == -1 ||
	    segtab(lo) == -1  ||
 	    modtab(lo) == -1  ||
 	    globtab(lo) == -1) {
 	    	lmf_close(lo);
 		return 0;
	}
	return lo;
}

lmf_close(lmfo *lo)
{
	CHKMAGIC(lo);
	if (lo->langtab) free(lo->langtab);
	if (lo->modtab)  free(lo->modtab);
	if (lo->globtab) free(lo->globtab);
	if (lo->segtab)  free(lo->segtab);
	if (lo->dbase)   free(lo->dbase);
	fclose(lo->fp);
	free(lo);
	return 0;
}

void *
lmf_demand(lmfo *lo, demand_info *dem, unsigned index, unsigned *size)
{
	char          *sbase = (char *)lo->sect;
	ulong_t        a;

	if (index > dem->num_entries)
		return 0;

	a = ((ulong_t*)(sbase+dem->offset))[index];
	*size = ((ulong_t*)(sbase+dem->offset))[index+1] - a;	
#ifdef DEBUGGING
printf("demand offset 0x%x->0x%x size %u\n", dem->offset, a, *size);
#endif
	return sbase + a;
}






